Redis replication stream processor storing data in TS (Rust-based storage).
This module provides Redis replication stream processing using TS (Rust-based storage) for efficient key-value storage. Currently supports sets, sorted sets, lists, hashes and strings.